import vipy.data.activitynet
import vipy.data.aflw
import vipy.data.ava
import vipy.data.caltech101
import vipy.data.caltech256
import vipy.data.casia
import vipy.data.charades
import vipy.data.cifar
import vipy.data.coco
import vipy.data.coil100
import vipy.data.d2d
import vipy.data.ethzshapes
import vipy.data.eurosat
import vipy.data.facescrub
import vipy.data.fddb
import vipy.data.food101
import vipy.data.hf
import vipy.data.hmdb
import vipy.data.imagenet
import vipy.data.inaturalist
import vipy.data.kinetics
import vipy.data.kthactions
import vipy.data.lfw
import vipy.data.megaface
import vipy.data.meva
import vipy.data.mit67
import vipy.data.mnist
import vipy.data.momentsintime
import vipy.data.msceleb
import vipy.data.objectnet
import vipy.data.oxford_flowers_102
import vipy.data.places
import vipy.data.stanford_dogs
import vipy.data.stanford_cars
import vipy.data.vggface
import vipy.data.vggface2
import vipy.data.visualgenome
import vipy.data.youtubefaces
import vipy.data.openimages
import vipy.data.ucf101
import vipy.data.lvis
import vipy.data.widerface
import vipy.data.kitti
import vipy.data.youtubeBB
import vipy.data.pip
import vipy.data.cap
import vipy.data.tiny_virat

